import numpy as np
import openmc

import pandas as pd
from omegah2csg import OmegaHMesh
from omegah2csg.convert2openmc import create_openmc_universe
from omegah2csg import read_edge_coefficients_from_file
from omegah2csg import read_face_connectivity_from_file
from omegah2csg.OmegaHMesh import EdgeType

from pathlib import Path


parent_directory = Path(__file__).resolve().parent


def test_read_from_file():
    edge_coefficient_filename = (
        parent_directory / "assets/6elem_mesh_files/edge-coefficients.dat"
    )
    [edge_coefficients_file, boundary_edge_ids_file] = read_edge_coefficients_from_file(
        edge_coefficient_filename
    )

    face_connectivity_filename = (
        parent_directory / "assets/6elem_mesh_files/face-edge-connectivity.dat"
    )
    face_connectivity_file = read_face_connectivity_from_file(
        face_connectivity_filename
    )

    with OmegaHMesh(parent_directory / "assets/6elem.osh") as mesh:
        [
            edge_coefficients_api,
            edge_types_,
            boundary_edge_ids_api,
            face_connectivity_api,
        ] = mesh.get_all_geometry_info()

    # compare boundary edges
    assert boundary_edge_ids_api.ndim == 1
    assert boundary_edge_ids_file.ndim == 1
    # 8 boundary edges
    assert boundary_edge_ids_api.shape[0] == boundary_edge_ids_file.shape[0] == 8

    # edge ids are the same
    for i in range(boundary_edge_ids_api.shape[0]):
        assert boundary_edge_ids_api[i] == boundary_edge_ids_file[i]

    assert edge_coefficients_api.ndim == edge_coefficients_file.ndim == 2
    assert edge_coefficients_api.shape[0] == edge_coefficients_file.shape[0] == 13
    assert edge_coefficients_api.shape[1] == edge_coefficients_file.shape[1] == 6

    # first 5 coefficients are the same
    for i in range(edge_coefficients_api.shape[0]):
        for j in range(edge_coefficients_file.shape[1] - 1):
            assert np.isclose(edge_coefficients_api[i, j], edge_coefficients_file[i, j])

    # check face connectivity
    assert face_connectivity_file.shape[0] == face_connectivity_api.shape[0] == 6
    assert face_connectivity_file.shape[1] == face_connectivity_api.shape[1] == 6

    for i in range(face_connectivity_file.shape[0]):
        for j in range(face_connectivity_file.shape[1]):
            assert face_connectivity_api[i, j] == face_connectivity_file[i, j]


def test_all_gemetry_info_6element_mesh():
    with OmegaHMesh(parent_directory / "assets/6elem.osh") as mesh:
        [edge_coefficients, edge_types, boundary_edge_ids, face_connctivity] = (
            mesh.get_all_geometry_info()
        )
        print(pd.DataFrame(edge_coefficients))
        print(pd.DataFrame(boundary_edge_ids))
        print(pd.DataFrame(face_connctivity))

    assert edge_coefficients.shape[0] == 13
    assert edge_coefficients.shape[1] == 6

    assert boundary_edge_ids.shape[0] == 8

    assert face_connctivity.shape[0] == 6
    assert face_connctivity.shape[1] == 6

    n_cones = 10
    n_cylinders = 3
    n_planes = 0
    assert np.sum(edge_types == EdgeType.Z_CONE) == n_cones
    assert np.sum(edge_types == EdgeType.Z_CYLINDER) == n_cylinders
    assert np.sum(edge_types == EdgeType.Z_PLANE) == n_planes


def test_all_gemetry_info_16_element_mesh():
    with OmegaHMesh(parent_directory / "assets/16elem.osh") as mesh:
        [edge_coefficients, edge_types, boundary_edge_ids, face_connctivity] = (
            mesh.get_all_geometry_info(tol=1e-4)
        )
        print(pd.DataFrame(edge_coefficients))
        print(pd.DataFrame(boundary_edge_ids))
        print(pd.DataFrame(face_connctivity))

    assert edge_coefficients.shape[0] == 30
    assert edge_coefficients.shape[1] == 6

    assert boundary_edge_ids.shape[0] == 12

    assert face_connctivity.shape[0] == 16
    assert face_connctivity.shape[1] == 6

    n_cones = 24
    n_cylinders = 4
    n_planes = 2
    assert n_cones + n_cylinders + n_planes == edge_coefficients.shape[0]
    assert np.sum(edge_types == EdgeType.Z_CONE) == n_cones
    assert np.sum(edge_types == EdgeType.Z_CYLINDER) == n_cylinders
    assert np.sum(edge_types == EdgeType.Z_PLANE) == n_planes


def test_create_openmc_universe():
    with OmegaHMesh(parent_directory / "assets/6elem.osh") as mesh:
        universe = create_openmc_universe(mesh)
        geom = openmc.Geometry(universe)
        geom.export_to_xml()


def test_edge_and_face_coefficients():
    tol = 1e-10
    with OmegaHMesh(parent_directory / "assets/6elem.osh") as mesh:
        edge_coefficients, edge_types = mesh.get_edge_coefficients(tol=tol)
        boundary_edge_ids = mesh.get_boundary_edge_ids()
        n_faces = mesh.num_entities(2)
        n_edges = mesh.num_entities(1)

        print("\n\nEdge coefficients:")
        print(pd.DataFrame(edge_coefficients))

        print("\n\nFace connectivity:")
        face_connectivity_edge_given = mesh.get_face_connectivity(
            edge_coefficients, tol=tol
        )
        print(pd.DataFrame(face_connectivity_edge_given))
        face_connectivity_edge_not_given = mesh.get_face_connectivity(tol=tol)

    top_bottom_flag = edge_coefficients[:, 5]
    edges = []
    intersections = edge_coefficients[:, 3]
    m2 = edge_coefficients[:, 1]
    z2 = edge_coefficients[:, 2]
    neg_c = edge_coefficients[:, 3]

    for i in range(len(intersections)):
        if edge_types[i] == EdgeType.Z_PLANE:
            edges.append(openmc.ZPlane(z0=-neg_c[i]))
        elif edge_types[i] == EdgeType.Z_CYLINDER:
            edges.append(openmc.ZCylinder(r=np.sqrt(-neg_c[i])))
        elif edge_types[i] == EdgeType.Z_CONE:
            edges.append(
                openmc.model.ZConeOneSided(
                    z0=intersections[i],
                    r2=1.0 / abs(m2[i]),
                    up=True if top_bottom_flag[i] == 1 else False,
                )
            )
        else:
            raise RuntimeError(
                f"Z2 value {z2[i]} not recognized. Coefficients: {edge_coefficients[i, :]}"
            )

    assert edge_coefficients.shape[0] == 13
    assert n_edges == 13
    assert edge_coefficients.shape[1] == 6
    assert boundary_edge_ids.shape[0] == 8

    print(f"\n\nBoundary edge ids: {boundary_edge_ids}")

    # assert that both face connectivities are the same
    assert face_connectivity_edge_given.shape[0] == n_faces
    assert face_connectivity_edge_given.shape[1] == 6

    assert (
        face_connectivity_edge_given.shape[0]
        == face_connectivity_edge_not_given.shape[0]
    )
    assert (
        face_connectivity_edge_given.shape[1]
        == face_connectivity_edge_not_given.shape[1]
    )

    for i in range(face_connectivity_edge_given.shape[0]):
        for j in range(face_connectivity_edge_given.shape[1]):
            assert (
                face_connectivity_edge_given[i, j]
                == face_connectivity_edge_not_given[i, j]
            )



"""
This module provides functions for converting Omega_h meshes to OpenMC geometry.

.. code-block:: python

    convert2openmc.create_openmc_geometry(mesh)
"""

import numpy as np
import openmc
from typing import Tuple

from ..OmegaHMesh import EdgeType, OmegaHMesh
from ..config import kokkos_runtime


Coord = Tuple[float, float]


def is_horizontal(p1: Coord, p2: Coord, tol=1e-6) -> bool:
    return np.abs(p1[1] - p2[1]) < tol


def is_vertical(p1: Coord, p2: Coord, tol=1e-6) -> bool:
    return np.abs(p1[0] - p2[0]) < tol


def get_slope(p1: Coord, p2: Coord, tol=1e-6):
    if is_horizontal(p1, p2, tol):
        return 0
    if is_vertical(p1, p2, tol):
        return np.nan

    m = (p2[1] - p1[1]) / (p2[0] - p1[0])
    return m


def get_vertical_intersection(p1: Coord, p2: Coord, tol=1e-6):
    if is_horizontal(p1, p2, tol):
        return p1[1]
    if is_vertical(p1, p2, tol):
        return np.nan

    m = get_slope(p1, p2, tol)
    c = p1[1] - m * p1[0]
    return c


def get_line_equation(p1: Coord, p2: Coord, tol=1e-6) -> Tuple:
    """
    Returns m,c for a line equation
    :param p1: Point 1 Coordinates
    :param p2: Point 2 Coordinates
    :param tol: Tolerance
    :return: m, c, up: Slope, Intersection, Up Flag for cone
    """
    if is_horizontal(p1, p2, tol):
        return 0, p1[1], np.nan
    if is_vertical(p1, p2, tol):
        return np.inf, np.nan, np.nan

    m = get_slope(p1, p2, tol)
    c = p1[1] - m * p1[0]

    # see doc for figure
    up = p1[1] > c and p2[1] > c
    # both should work since both points are on the right of z
    assert up == (p1[1] > c or p2[1] > c)
    return m, c, up


def is_cone(p1: Coord, p2: Coord, tol=1e-6) -> bool:
    return not is_horizontal(p1, p2, tol) and not is_vertical(p1, p2, tol)


def create_z_cone(p1: Coord, p2: Coord, tol=1e-6) -> openmc.model.ZConeOneSided:
    assert not is_horizontal(p1, p2, tol), (
        "Horizonal line. Use create_z_plane instead of create_z_cone."
    )
    assert not is_vertical(p1, p2, tol), (
        "Vertical line. Use create_z_cylinder instead of create_z_cone."
    )

    m, c, up = get_line_equation(p1, p2, tol)
    assert not np.isinf(m), "Slope m found to be infinity."
    assert not np.isnan(c), "Slope c found to be nan."

    cone = openmc.model.ZConeOneSided(x0=0, y0=0, z0=c, r2=m * m, up=up)
    return cone


def create_z_plane(p1: Coord, p2: Coord, tol=1e-6) -> openmc.ZPlane:
    assert is_horizontal(p1, p2, tol), "Not a horizontal line"
    return openmc.ZPlane(z0=p1[1])


def create_z_cylinder(p1: Coord, p2: Coord, tol=1e-6) -> openmc.ZCylinder:
    assert is_vertical(p1, p2, tol), "Not a vertical line"
    return openmc.ZCylinder(r=p1[0])


def create_openmc_surface(p1: Coord, p2: Coord, tol=1e-6):
    if is_horizontal(p1, p2, tol):
        return create_z_plane(p1, p2, tol)
    if is_vertical(p1, p2, tol):
        return create_z_cylinder(p1, p2, tol)
    if is_cone(p1, p2, tol):
        return create_z_cone(p1, p2, tol)

    raise RuntimeError(f"Error creating surface: {p1} and {p2}")


def create_openmc_universe(
    mesh: OmegaHMesh, materials=None, print_debug=False, tol=1e-6
):
    """Create OpenMC geometry (OpenMC.Universe) from OmegaHMesh by rorating along Z-axis

    Parameters
    ----------
    mesh: OmegaHMesh
        OmegaHMesh object with isOnWall and offset_face tags
    materials: Any interable of size of number of cells, optional
        Optional array of OpenMC materials to fill the cells. Size should match number of faces in the mesh.
        If None, dummy materials will be used to avoid OpenMC errors.
    print_debug: boo, optional
        If True, prints edge coefficients and face connectivity for debugging.
    tol: float, optional
        Tolerance for determining edge types and connectivity. Should be a small positive number, e.g. 1e-6.

    Returns
    -------
    OpenMC.Universe
        OpenMC Universe object representing the geometry

    """
    if not kokkos_runtime.is_running():
        raise RuntimeError("Kokkos not running...")

    [edge_coefficients, edge_types, boundary_edge_ids, face_connctivity] = (
        mesh.get_all_geometry_info(print_debug, tol)
    )
    n_edges = mesh.num_entities(1)
    n_faces = mesh.num_entities(2)

    top_bottom_flag = edge_coefficients[:, 4]
    edges = np.empty(shape=n_edges, dtype=object)
    intersections = edge_coefficients[:, 2] / 2.0
    m2 = edge_coefficients[:, 0]
    z2 = edge_coefficients[:, 1]
    neg_c = edge_coefficients[:, 3]

    for i in range(len(intersections)):
        if edge_types[i] == EdgeType.Z_PLANE:
            edges[i] = openmc.ZPlane(z0=-neg_c[i])
        elif edge_types[i] == EdgeType.Z_CYLINDER:
            edges[i] = openmc.ZCylinder(r=np.sqrt(-neg_c[i]))
        elif edge_types[i] == EdgeType.Z_CONE:
            edges[i] = openmc.model.ZConeOneSided(
                z0=intersections[i],
                r2=1.0 / abs(m2[i]),
                up=True if top_bottom_flag[i] == 1 else False,
            )
        else:
            raise RuntimeError(
                f"Z2 value {z2[i]} not recognized. Coefficients: {edge_coefficients[i, :]}"
            )

    for edge_id in boundary_edge_ids:
        edges[edge_id].boundary_type = "reflective"

    if materials is None:
        # to stop openmc errors
        dummy = openmc.Material(name="dummy")
        dummy.add_nuclide("H1", 1.0)
        dummy.set_density(units="g/cc", density=1.0)
        materials = np.array([dummy] * n_faces)

    assert materials.ndim == 1
    assert materials.shape[0] == n_faces

    cells = np.empty(shape=n_faces, dtype=object)
    for i in range(face_connctivity.shape[0]):
        vol1 = (
            +edges[int(face_connctivity[i][0])]
            if int(face_connctivity[i][1]) == 1
            else -edges[int(face_connctivity[i][0])]
        )
        vol2 = (
            +edges[int(face_connctivity[i][2])]
            if int(face_connctivity[i][3]) == 1
            else -edges[int(face_connctivity[i][2])]
        )
        vol3 = (
            +edges[int(face_connctivity[i][4])]
            if int(face_connctivity[i][5]) == 1
            else -edges[int(face_connctivity[i][4])]
        )

        cells[i] = openmc.Cell(
            region=vol1 & vol2 & vol3,
            fill=materials[i],
            name="cell" + str(i),
            cell_id=i,
        )

    universe = openmc.Universe(cells=cells)
    return universe


def convert2openmcXML(filename, tol):
    assert filename.endswith(".osh")
    assert (tol < 1e-6) and (tol > 0.0)

    with OmegaHMesh(filename) as mesh:
        universe = create_openmc_universe(mesh=mesh, tol=tol)
        geom = openmc.Geometry(universe)
        geom.export_to_xml()



from .convert2openmc import (
    is_horizontal,
    is_vertical,
    get_slope,
    get_vertical_intersection,
    get_line_equation,
    is_cone,
    create_z_cone,
    create_z_plane,
    create_z_cylinder,
    create_openmc_surface,
    create_openmc_universe,
    convert2openmcXML,
)

__all__ = [
    "is_horizontal",
    "is_vertical",
    "get_slope",
    "get_vertical_intersection",
    "get_line_equation",
    "is_cone",
    "create_z_cone",
    "create_z_plane",
    "create_z_cylinder",
    "create_openmc_surface",
    "create_openmc_universe",
    "convert2openmcXML",
]



import numpy as np
import openmc
from omegah2csg.convert2openmc import get_line_equation
from omegah2csg.convert2openmc import create_openmc_surface
from omegah2csg.convert2openmc import is_horizontal, is_vertical, is_cone


def test_is_horizontal_vertical_cone():
    p1 = (1, 1)
    p2 = (2, 3)

    assert not is_horizontal(p1, p2)
    assert not is_vertical(p1, p2)
    assert is_cone(p1, p2)

    p1 = (1, 1)
    p3 = (1.0, -10.0)
    assert not is_horizontal(p1, p3)
    assert is_vertical(p1, p3)
    assert not is_cone(p1, p3)

    p2 = (2, 3)
    p4 = (-1.0, 3)
    assert is_horizontal(p2, p4)
    assert not is_vertical(p2, p4)
    assert not is_cone(p2, p4)


def test_get_line_equation():
    p1 = (1, 1)
    p2 = (2, 2)

    m, c, up = get_line_equation(p1, p2)
    assert np.isclose(m, 1)
    assert np.isclose(c, 0)
    assert up


def test_create_surface_cone():
    p1 = (1, 1)
    p2 = (2, 2)

    cone = create_openmc_surface(p1, p2, 1e-6)
    assert type(cone) is openmc.model.ZConeOneSided
    disambiguation_surface = cone.plane
    assert np.isclose(disambiguation_surface.z0, 0.0)
    assert cone.up

    p1 = (1, 1)
    p2 = (2, 3)

    cone = create_openmc_surface(p1, p2, 1e-6)
    assert type(cone) is openmc.model.ZConeOneSided
    disambiguation_surface = cone.plane
    assert np.isclose(disambiguation_surface.z0, -1.0)
    assert cone.up

    p1 = (1, 1)
    p2 = (2, -1)

    cone = create_openmc_surface(p1, p2, 1e-6)
    assert type(cone) is openmc.model.ZConeOneSided
    disambiguation_surface = cone.plane
    assert np.isclose(disambiguation_surface.z0, 3.0)
    assert not cone.up



import argparse

from .convert2degas2 import convert2degas2


def main():
    parser = argparse.ArgumentParser(
        description="Convert Omega_h mesh to DEGAS2 geometry (netcdf file)"
    )
    parser.add_argument("filename", help="Omega_h mesh file (.osh)")
    parser.add_argument("--tol", type=float, help="tolerance", default=1e-10)
    parser.add_argument(
        "--auxfiles",
        type=bool,
        help="create dummy auxiliary files for degas2 case",
        default=False,
    )
    args = parser.parse_args()

    convert2degas2(
        mesh_filename=args.filename,
        tol=args.tol,
        netcdf_filename="geometry.nc",
        create_aux_files=args.auxfiles,
    )


# For setuptools entry point
app = main

if __name__ == "__main__":
    main()



from omegah2csg import OmegaHMesh
from pathlib import Path

parent_directory = Path(__file__).resolve().parent


def test_mesh_loading():
    with OmegaHMesh(parent_directory / "assets/6elem.osh") as mesh:
        assert mesh.dim == 2
        assert mesh.num_entities(mesh.dim) == 6


def test_multiple_mesh_loading():
    with OmegaHMesh(parent_directory / "assets/16elem.osh") as mesh:
        assert mesh.dim == 2
        assert mesh.num_entities(mesh.dim) == 16


def test_get_node_coordinates():
    with OmegaHMesh(parent_directory / "assets/6elem.osh") as mesh:
        node_coordinates = mesh.get_node_coordinates()
        n_nodes = mesh.num_entities(0)
        assert node_coordinates.shape[0] == n_nodes * 2

        print("\nNode coordinates:")
        for i in range(n_nodes):
            print(
                f"Node {i}: {float(node_coordinates[i * 2]), float(node_coordinates[i * 2 + 1])}"
            )



from ._lib import _dll


class KokkosRuntime:
    _kokkos_initialized = False
    _kokkos_finalized = False

    # ============= Kokkos ====================

    @property
    def kokkos_initialized(self):
        return self._kokkos_initialized

    @property
    def kokkos_finalized(self):
        return self._kokkos_finalized

    def kokkos_initialize(self):
        """Call before any operations are performed."""
        if self._kokkos_finalized:
            raise RuntimeError(
                "Kokkos already finalized. Restart python to initialize kokkos again."
            )

        if self._kokkos_initialized:
            raise RuntimeWarning("Kokkos already initialized. Doing nothing.")

        _dll.kokkos_initialize()
        self._kokkos_initialized = True

    def kokkos_finalize(self):
        """Call after all operations are performed."""
        if self._kokkos_finalized:
            raise RuntimeWarning(
                "Kokkos already finalized. Restart python to initialize kokkos again."
            )

        _dll.kokkos_finalize()
        self._kokkos_finalized = True

    def is_running(self):
        return self._kokkos_initialized and (not self._kokkos_finalized)


kokkos_runtime = KokkosRuntime()



import argparse

from .convert2openmc import convert2openmcXML


def main():
    parser = argparse.ArgumentParser(
        description="Convert Omega_h mesh to OpenMC geometry"
    )
    parser.add_argument("filename", help="Omega_h mesh file (.osh)")
    parser.add_argument("--tol", type=float, help="tolerance", default=1e-10)
    args = parser.parse_args()

    convert2openmcXML(filename=args.filename, tol=args.tol)


# For setuptools entry point
app = main

if __name__ == "__main__":
    main()



from ctypes import c_void_p, c_char_p, Structure, c_int, c_bool, c_double

import numpy as np
from numpy.ctypeslib import ndpointer

from ._lib import _dll
from .config import kokkos_runtime
import os
from enum import IntEnum


# Define C structures
class OmegaHLibraryPointer(Structure):
    _fields_ = [("pointer", c_void_p)]


class OmegaHMeshPointer(Structure):
    _fields_ = [("pointer", c_void_p)]


class EdgeType(IntEnum):
    Z_PLANE = 1
    Z_CYLINDER = 2
    Z_CONE = 3


# Define function prototypes
_dll.create_omegah_library.restype = OmegaHLibraryPointer
_dll.create_omegah_library.argtypes = []

_dll.destroy_omegah_library.argtypes = [OmegaHLibraryPointer]

_dll.create_omegah_mesh.restype = OmegaHMeshPointer
_dll.create_omegah_mesh.argtypes = [OmegaHLibraryPointer, c_char_p]

_dll.destroy_omegah_mesh.argtypes = [OmegaHMeshPointer]

_dll.print_mesh_info.argtypes = [OmegaHMeshPointer]

_dll.get_num_entities.restype = c_int
_dll.get_num_entities.argtypes = [OmegaHMeshPointer, c_int]

_dll.get_dim.restype = c_int
_dll.get_dim.argtypes = [OmegaHMeshPointer]

_dll.capi_is_mesh_bounded_by_box.restype = c_bool
_dll.capi_is_mesh_bounded_by_box.argtypes = [OmegaHMeshPointer]

_dll.capi_has_boundary_layer.restype = c_bool
_dll.capi_has_boundary_layer.argtypes = [OmegaHMeshPointer]

_dll.capi_get_mesh_int_tag_array.restype = c_bool
_dll.capi_get_mesh_int_tag_array.argtypes = [
    OmegaHMeshPointer,
    c_int,
    c_char_p,
    ndpointer(c_int),
    c_int,
]

_dll.capi_get_cell_bounding_boxes.argtypes = [
    OmegaHMeshPointer,
    ndpointer(c_double),
    c_int,
]

_dll.capi_get_cell_volumes.argtypes = [OmegaHMeshPointer, ndpointer(c_double), c_int]

_dll.capi_get_cell_centroids.argtypes = [OmegaHMeshPointer, ndpointer(c_double), c_int]

_dll.capi_get_edge_coordinates.argtypes = [
    OmegaHMeshPointer,
    ndpointer(c_double),
    c_int,
]

_dll.capi_get_number_of_edges_inside_wall.argtypes = [OmegaHMeshPointer]
_dll.capi_get_number_of_edges_inside_wall.restype = c_int

_dll.capi_get_edge_to_face_connectivity.argtypes = [
    OmegaHMeshPointer,
    ndpointer(c_int),
    c_int,
]

_dll.capi_get_wall_adjacent_triangles.argtypes = [
    OmegaHMeshPointer,
    ndpointer(c_int),
    c_int,
]

_dll.capi_get_wall_edge_ids.argtypes = [OmegaHMeshPointer, ndpointer(c_int), c_int]

_dll.capi_get_edge_to_face_connectivity.argtypes = [
    OmegaHMeshPointer,
    ndpointer(c_int),
    c_int,
]

_dll.capi_get_node_coordinates.argtypes = [
    OmegaHMeshPointer,
    ndpointer(c_double),
    c_int,
]

_dll.capi_compute_edge_coefficients.restype = None
_dll.capi_compute_edge_coefficients.argtypes = [
    OmegaHMeshPointer,
    c_int,
    ndpointer(c_double),
    ndpointer(c_int),
    c_bool,
    c_double,
]

_dll.capi_get_boundary_edge_ids.restype = None
_dll.capi_get_boundary_edge_ids.argtypes = [OmegaHMeshPointer, c_int, ndpointer(c_int)]

_dll.capi_get_number_of_boundary_edges.restype = c_int
_dll.capi_get_number_of_boundary_edges.argtypes = [OmegaHMeshPointer]

_dll.capi_get_face_connectivity.restype = None
_dll.capi_get_face_connectivity.argtypes = [
    OmegaHMeshPointer,
    c_int,
    ndpointer(c_double),
    c_int,
    ndpointer(c_int),
    c_bool,
    c_double,
]

_dll.capi_get_all_geometry_info.restype = None
_dll.capi_get_all_geometry_info.argtypes = [
    OmegaHMeshPointer,
    c_int,
    c_int,
    ndpointer(c_double),
    ndpointer(c_int),
    ndpointer(c_int),
    ndpointer(c_int),
    c_bool,
    c_double,
]


class OmegaHMesh:
    """
    A context manager for loading and working with OmegaH meshes.

    Example usage:
     with OmegaHMesh(filename) as mesh:
         mesh = OmegaHMesh()
    """

    def __init__(self, filename):
        self.filename = str(filename)
        self.lib = None
        self.mesh = None
        if not kokkos_runtime.kokkos_initialized:
            kokkos_runtime.kokkos_initialize()

        # check if given directory exits
        if not os.path.isdir(filename):
            raise RuntimeError(f"OmegaHMesh filename {filename} does not exist.")

    def __enter__(self):
        # Create library
        self.lib = _dll.create_omegah_library()

        # Create mesh in try catch to capture error
        try:
            self.mesh = _dll.create_omegah_mesh(self.lib, self.filename.encode("utf-8"))
        except Exception as exception:
            raise RuntimeError(f"Error creating OmegaH mesh: {exception}")

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Clean up in reverse order
        if self.mesh:
            _dll.destroy_omegah_mesh(self.mesh)
        if self.lib:
            _dll.destroy_omegah_library(self.lib)

    def print_info(self):
        """Print information about the loaded mesh."""
        if self.mesh is None:
            raise RuntimeError("No mesh loaded. Use as context manager.")
        print(f"Mesh information for: {self.filename}")
        _dll.print_mesh_info(self.mesh)

    def num_entities(self, dim):
        """Number of entities of given dimension."""
        if self.mesh is None:
            raise RuntimeError("No mesh loaded. Use as context manage.")
        return _dll.get_num_entities(self.mesh, dim)

    @property
    def dim(self):
        """Dimension of given mesh."""
        if self.mesh is None:
            raise RuntimeError("No mesh loaded. Use as context manager.")
        return _dll.get_dim(self.mesh)

    def get_node_coordinates(self) -> np.ndarray:
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")

        try:
            num_nodes = self.num_entities(0)
            size = 2 * num_nodes
            coordinates = np.empty(size, dtype=np.float64)
            _dll.capi_get_node_coordinates(self.mesh, coordinates, size)
            return coordinates
        except Exception as exception:
            raise RuntimeError(f"Error getting node coordinates: {exception}")

    def is_bounded_by_box(self) -> bool:
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")

        try:
            return _dll.capi_is_mesh_bounded_by_box(self.mesh)
        except Exception as exception:
            raise RuntimeError(f"Error finding box: {exception}")

    def has_boundary_layer(self) -> bool:
        """
        Checks if the mesh has a single boundary layer after the wall
        :return: If the input mesh has a single boundary layer after the wall
        """
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")

        try:
            return _dll.capi_has_boundary_layer(self.mesh)
        except Exception as exception:
            raise RuntimeError(f"Error finding layer: {exception}")

    def get_integer_tag_array(self, dim, name) -> np.ndarray:
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")

        try:
            array_size = self.num_entities(dim)
            tag_array = np.empty(array_size, dtype=np.int32)
            success = _dll.capi_get_mesh_int_tag_array(
                self.mesh, dim, str(name).encode("utf-8"), tag_array, array_size
            )
            assert success, "Internal error"
            return tag_array
        except Exception as exception:
            raise RuntimeError(f"Error getting integer tag: {exception}")

    def get_boundary_face_flag(self) -> np.ndarray:
        return self.get_integer_tag_array(2, "offset_face")

    def get_cell_bounding_boxes(self) -> np.ndarray:
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")
        try:
            n_cells = self.num_entities(2)
            cell_bounding_boxes = np.empty(n_cells * 4, dtype=np.float64)
            _dll.capi_get_cell_bounding_boxes(
                self.mesh, cell_bounding_boxes, n_cells * 4
            )
            return cell_bounding_boxes
        except Exception as exception:
            raise RuntimeError(f"Error getting cell bounding boxes: {exception}")

    def get_cell_volumes(self) -> np.ndarray:
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")

        try:
            n_cells = self.num_entities(2)
            cell_volumes = np.empty(n_cells, dtype=np.float64)
            _dll.capi_get_cell_volumes(self.mesh, cell_volumes, n_cells)
            return cell_volumes
        except Exception as exception:
            raise RuntimeError(f"Error getting cell volumes: {exception}")

    def get_cell_centroids(self) -> np.ndarray:
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")

        try:
            n_cells = self.num_entities(2)
            cell_centroids = np.empty(2 * n_cells, dtype=np.float64)
            _dll.capi_get_cell_centroids(self.mesh, cell_centroids, 2 * n_cells)
            return cell_centroids
        except Exception as exception:
            raise RuntimeError(f"Error getting cell centroids: {exception}")

    def get_edge_coordinates(self) -> np.ndarray:
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")

        try:
            n_edges = self.num_entities(1)
            edge_coordinates = np.empty(n_edges * 4, dtype=np.float64)
            _dll.capi_get_edge_coordinates(self.mesh, edge_coordinates, n_edges * 4)
            return edge_coordinates
        except Exception as exception:
            raise RuntimeError(f"Error getting edge coordinates: {exception}")

    def get_number_of_edges_inside_wall(self) -> int:
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")

        try:
            return _dll.capi_get_number_of_edges_inside_wall(self.mesh)
        except Exception as exception:
            raise RuntimeError(f"Error getting number of edges: {exception}")

    def get_edge_to_face_map(self) -> np.ndarray:
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")

        try:
            nedges = self.num_entities(1)
            edge_map_size = nedges * 2
            edge_to_face_map = np.empty(edge_map_size, dtype=np.int32)
            _dll.capi_get_edge_to_face_connectivity(
                self.mesh, edge_to_face_map, edge_map_size
            )
            return edge_to_face_map.reshape((nedges, 2))

        except Exception as exception:
            raise RuntimeError(f"Error getting edge to face map: {exception}")

    def get_wall_edge_ids(self) -> np.ndarray:
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")

        num_first_wall_points = np.sum(self.get_integer_tag_array(0, "isOnWall"))
        wall_edge_ids = np.empty(num_first_wall_points, dtype=np.int32)

        try:
            _dll.capi_get_wall_edge_ids(
                self.mesh, wall_edge_ids, wall_edge_ids.shape[0]
            )
            return wall_edge_ids
        except Exception as exception:
            raise RuntimeError(f"Error getting wall edge ids: {exception}")

    def get_wall_adjacent_triangles(self) -> np.ndarray:
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")

        num_first_wall_points = np.sum(self.get_integer_tag_array(0, "isOnWall"))
        wall_adjacent_triangles = np.empty(2 * num_first_wall_points, dtype=np.int32)

        try:
            _dll.capi_get_wall_adjacent_triangles(
                self.mesh, wall_adjacent_triangles, 2 * num_first_wall_points
            )
            return wall_adjacent_triangles
        except Exception as exception:
            raise RuntimeError(f"Error getting wall adjacent triangles: {exception}")

    def get_edge_coefficients(self, print_debug=False, tol=1e-6):
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")

        num_edges = self.num_entities(1)
        coefficients = np.zeros(num_edges * 6, dtype=np.float64)
        edge_types = np.empty(num_edges, dtype=np.int32)
        size = coefficients.size

        try:
            _dll.capi_compute_edge_coefficients(
                self.mesh, c_int(size), coefficients, edge_types, print_debug, tol
            )

        except Exception as exception:
            raise RuntimeError(f"Error computing edge coefficients: {exception}")

        return coefficients.reshape(num_edges, 6), edge_types

    def get_num_of_boundary_edges(self) -> int:
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")

        try:
            return _dll.capi_get_number_of_boundary_edges(self.mesh)

        except Exception as exception:
            raise RuntimeError(f"Error computing number of boundary edges: {exception}")

    def get_boundary_edge_ids(self, size=0):
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")

        if size == 0:
            size = self.get_num_of_boundary_edges()

        boundary_edge_ids = np.zeros(size, dtype=np.int32)
        try:
            _dll.capi_get_boundary_edge_ids(self.mesh, size, boundary_edge_ids)
        except Exception as exception:
            raise RuntimeError(f"Error computing boundary edge ids: {exception}")

        return boundary_edge_ids

    def get_face_connectivity(
        self, edge_coefficients=None, print_debug=False, tol=1e-6
    ):
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")

        n_faces = self.num_entities(2)
        n_edges = self.num_entities(1)

        if edge_coefficients is None:
            edge_coefficients, edge_types_ = self.get_edge_coefficients()

        assert edge_coefficients.shape[0] == n_edges
        assert edge_coefficients.shape[1] == 6
        assert edge_coefficients.dtype == np.float64

        edge_coefficients_shaped = edge_coefficients.reshape(n_edges * 6)

        face_connectivity = np.zeros(n_faces * 6, dtype=np.int32)

        # TODO add core-dump capture
        try:
            _dll.capi_get_face_connectivity(
                self.mesh,
                n_edges * 6,
                edge_coefficients_shaped,
                n_faces * 6,
                face_connectivity,
                print_debug,
                tol,
            )
        except Exception as exception:
            raise RuntimeError(f"Error computing face connectivity: {exception}")

        return face_connectivity.reshape(n_faces, 6)

    def get_all_geometry_info(self, print_debug=False, tol=1e-6):
        if not kokkos_runtime.is_running():
            raise RuntimeError("Kokkos not running...")

        n_faces = self.num_entities(2)
        n_edges = self.num_entities(1)

        edge_coefficients = np.zeros(n_edges * 6, dtype=np.float64)
        edge_types = np.empty(n_edges, dtype=np.int32)
        boundary_edge_ids_num = self.get_num_of_boundary_edges()
        boundary_edge_ids = np.zeros(boundary_edge_ids_num, dtype=np.int32)
        face_connctivity = np.zeros(n_faces * 6, dtype=np.int32)

        try:
            _dll.capi_get_all_geometry_info(
                self.mesh,
                n_edges,
                n_faces,
                edge_coefficients,
                edge_types,
                boundary_edge_ids,
                face_connctivity,
                print_debug,
                tol,
            )
        except Exception as exception:
            raise RuntimeError(f"Error computing all geometry info: {exception}")

        return (
            edge_coefficients.reshape(n_edges, 6),
            edge_types,
            boundary_edge_ids,
            face_connctivity.reshape(n_faces, 6),
        )


def read_edge_coefficients_from_file(filename):
    with open(filename, "r") as f:
        data = f.readlines()
        n_edges = len(data) - 2

        # this file has first 5 coefficients
        edge_coefficients = np.zeros((n_edges, 6), dtype=np.float64)

        for i in range(n_edges):
            values = data[i].split()
            assert len(values) == 6, f"Length of edge_coefficients is not 6: {values}"

            edge_id = int(values[0])
            assert edge_id == i, f"Edge id is not {i}, edge id {edge_id}"

            # fill the 5 coefficients
            edge_coefficients[edge_id, 0] = np.float64(values[1])
            edge_coefficients[edge_id, 1] = np.float64(values[2])
            edge_coefficients[edge_id, 2] = np.float64(values[3])
            edge_coefficients[edge_id, 3] = np.float64(values[4])
            edge_coefficients[edge_id, 4] = np.float64(values[5])

        boundary_edges_line = data[-2].split()
        assert len(boundary_edges_line) == 3
        num_boundary_edges = int(boundary_edges_line[-1])
        boundary_edge_ids = np.zeros(num_boundary_edges, dtype=np.int32)

        boundary_edge_ids_line = data[-1].split()
        assert len(boundary_edge_ids_line) == num_boundary_edges
        for i in range(num_boundary_edges):
            boundary_edge_ids[i] = int(boundary_edge_ids_line[i])

    return edge_coefficients, boundary_edge_ids


def read_face_connectivity_from_file(filename):
    with open(filename, "r") as f:
        data = f.readlines()
        n_faces = len(data)
        connectivity = np.zeros((n_faces, 6), dtype=np.int32)

        for i in range(n_faces):
            values = data[i].split()
            assert len(values) == 7
            face_id = int(values[0])
            assert face_id == i

            for j in range(6):
                connectivity[i, j] = np.int32(values[j + 1])

    return connectivity



from .config import KokkosRuntime, kokkos_runtime
from .OmegaHMesh import (
    OmegaHMesh,
    read_edge_coefficients_from_file,
    read_face_connectivity_from_file,
)
from .convert2openmc import (
    convert2openmcXML,
)
from .convert2degas2 import convert2degas2

__all__ = [
    # runtime
    "KokkosRuntime",
    "kokkos_runtime",
    # mesh
    "OmegaHMesh",
    "read_edge_coefficients_from_file",
    "read_face_connectivity_from_file",
    # openmc
    "convert2openmcXML",
    # degas2
    "convert2degas2",
]



from pathlib import Path
import numpy as np

from omegah2csg import OmegaHMesh
from omegah2csg import convert2degas2
import netCDF4 as nc

parent_dir = Path(__file__).resolve().parent
omega_h_mesh_to_convert = parent_dir / "assets/simple_degas2_case/tagged-dg2mesh.osh"
gold_nc_mesh = parent_dir / "assets/simple_degas2_case/gold-geometry.nc"
dg2d_converted_mesh = (
    parent_dir / "assets/simple_degas2_case/definegeometry2d-geometry.nc"
)


def test_is_bounded_by_box():
    with OmegaHMesh(omega_h_mesh_to_convert) as mesh:
        assert mesh.is_bounded_by_box(), "This mesh is bounded by box but found not"


def test_has_boundary_layer():
    with OmegaHMesh(omega_h_mesh_to_convert) as mesh:
        assert mesh.has_boundary_layer(), (
            "This mesh wrapped with a boundary layer but found false"
        )


def test_get_integer_tag_array():
    with OmegaHMesh(omega_h_mesh_to_convert) as mesh:
        offset_face_tag = mesh.get_integer_tag_array(2, "offset_face")
        num_offset_faces = offset_face_tag.sum()
        print(f"Number of offset faces {num_offset_faces}")
        assert num_offset_faces == 28, (
            "For this mesh, there are 28 offset faces."
            "Check picture in assets/simple_degas2_case/README.md"
        )


def test_get_cell_volumes():
    with OmegaHMesh(omega_h_mesh_to_convert) as mesh:
        boundary_face_flag = mesh.get_boundary_face_flag()
        num_boundary_faces = boundary_face_flag.sum()
        num_inside_faces = mesh.num_entities(2) - num_boundary_faces
        oh_cell_volumes = mesh.get_cell_volumes()
        oh_total_volume = np.sum(oh_cell_volumes)

    with nc.Dataset(gold_nc_mesh, "r", format="NETCDF4_CLASSIC") as mesh:
        gold_nc_zone_volumes = mesh["zone_volume"][:]
        gold_nc_total_volume = mesh["universal_cell_vol"][:]
        assert np.isclose(oh_total_volume, gold_nc_total_volume, rtol=1e-5), (
            f"Total volume mismatch. {oh_total_volume=} , {gold_nc_total_volume=}"
        )

    with nc.Dataset(dg2d_converted_mesh, "r", format="NETCDF4_CLASSIC") as mesh:
        dg2d_total_volume = mesh["universal_cell_vol"][:]
        dg2d_nc_zone_volumes = mesh["zone_volume"][:]
        assert np.isclose(oh_total_volume, dg2d_total_volume, rtol=1e-5), (
            f"Total volume mismatch. {oh_total_volume=} , {dg2d_total_volume=}"
        )

    for i in range(0, num_inside_faces):
        assert np.isclose(gold_nc_zone_volumes[i], oh_cell_volumes[i], rtol=1e-5), (
            f"{gold_nc_zone_volumes[i]} ≠ {oh_cell_volumes[i]} for face {i}"
        )
        assert np.isclose(dg2d_nc_zone_volumes[i], oh_cell_volumes[i], rtol=1e-5), (
            f"{dg2d_nc_zone_volumes[i]} ≠ {oh_cell_volumes[i]} for face {i}"
        )


def test_get_cell_bounding_boxes():
    with OmegaHMesh(omega_h_mesh_to_convert) as mesh:
        boundary_face_flag = mesh.get_boundary_face_flag()
        num_boundary_faces = boundary_face_flag.sum()
        num_inside_faces = mesh.num_entities(2) - num_boundary_faces
        oh_cell_bboxes = mesh.get_cell_bounding_boxes()

    with nc.Dataset(gold_nc_mesh, "r", format="NETCDF4_CLASSIC") as mesh:
        gold_nc_bbox_mins = mesh["zone_min"][:]
        gold_nc_bbox_maxs = mesh["zone_max"][:]

    with nc.Dataset(dg2d_converted_mesh, "r", format="NETCDF4_CLASSIC") as mesh:
        dg2d_nc_bbox_mins = mesh["zone_min"][:]
        dg2d_nc_bbox_maxs = mesh["zone_max"][:]

    for i in range(0, num_inside_faces):
        # xmin
        coord_tol = 1e-5
        assert np.isclose(
            oh_cell_bboxes[i * 4 + 0], gold_nc_bbox_mins[i, 0], atol=coord_tol
        ), f"Bounding box mismatch for face {i}"
        assert np.isclose(
            oh_cell_bboxes[i * 4 + 0], dg2d_nc_bbox_mins[i, 0], atol=coord_tol
        ), f"Bounding box mismatch for face {i}"
        # ymin
        assert np.isclose(
            oh_cell_bboxes[i * 4 + 1], gold_nc_bbox_mins[i, 2], atol=coord_tol
        ), f"Bounding box mismatch for face {i}"
        assert np.isclose(
            oh_cell_bboxes[i * 4 + 1], dg2d_nc_bbox_mins[i, 2], atol=coord_tol
        ), f"Bounding box mismatch for face {i}"
        # xmax
        assert np.isclose(
            oh_cell_bboxes[i * 4 + 2], gold_nc_bbox_maxs[i, 0], atol=coord_tol
        ), f"Bounding box mismatch for face {i}"
        assert np.isclose(
            oh_cell_bboxes[i * 4 + 2], dg2d_nc_bbox_maxs[i, 0], atol=coord_tol
        ), f"Bounding box mismatch for face {i}"
        # ymax
        assert np.isclose(
            oh_cell_bboxes[i * 4 + 3], gold_nc_bbox_maxs[i, 2], atol=coord_tol
        ), f"Bounding box mismatch for face {i}"
        assert np.isclose(
            oh_cell_bboxes[i * 4 + 3], dg2d_nc_bbox_maxs[i, 2], atol=coord_tol
        ), f"Bounding box mismatch for face {i}"


def test_get_cell_centroids():
    with OmegaHMesh(omega_h_mesh_to_convert) as mesh:
        boundary_face_flag = mesh.get_boundary_face_flag()
        num_boundary_faces = boundary_face_flag.sum()
        num_inside_faces = mesh.num_entities(2) - num_boundary_faces
        oh_cell_centroids = mesh.get_cell_centroids()

    with nc.Dataset(gold_nc_mesh, "r", format="NETCDF4_CLASSIC") as mesh:
        gold_zone_centroids = mesh["zone_center"][:]

    with nc.Dataset(dg2d_converted_mesh, "r", format="NETCDF4_CLASSIC") as mesh:
        dg2d_zone_centroids = mesh["zone_center"][:]

    for i in range(0, num_inside_faces):
        coord_tol = 1e-5
        assert np.isclose(
            oh_cell_centroids[i * 2 + 0], gold_zone_centroids[i, 0], atol=coord_tol
        ), f"Centroid x mismatch for face {i}"
        assert np.isclose(
            oh_cell_centroids[i * 2 + 0], dg2d_zone_centroids[i, 0], atol=coord_tol
        ), f"Centroid x mismatch for face {i}"

        assert np.isclose(
            oh_cell_centroids[i * 2 + 1], gold_zone_centroids[i, 2], atol=coord_tol
        ), f"Centroid y mismatch for face {i}"
        assert np.isclose(
            oh_cell_centroids[i * 2 + 1], dg2d_zone_centroids[i, 2], atol=coord_tol
        ), f"Centroid y mismatch for face {i}"


def test_get_number_of_edges_inside_wall():
    with OmegaHMesh(omega_h_mesh_to_convert) as mesh:
        number_of_edges_inside_wall = mesh.get_number_of_edges_inside_wall()
        assert number_of_edges_inside_wall == 14, (
            f"Number of edges_inside_wall (expected 14) = {number_of_edges_inside_wall}"
        )


def test_convert2degas2():
    convert2degas2(omega_h_mesh_to_convert, create_aux_files=False)


def test_small_netcdf_write():
    """
    The tests if the netcdf installation (python) is working on the current platform.
    """
    root_g = nc.Dataset("test.nc", mode="w", format="NETCDF4")
    root_g.createDimension("vector", 3)
    integer_scalar_var = root_g.createVariable("int_scalar", "i4", ())

    vector_var = root_g.createVariable("vector_var", "f8", ("vector",))
    print(f"{vector_var.shape=}")
    some_vector = np.array([0.0, 0.0, 0.0])
    print(f"{some_vector.shape=}")
    vector_var[:] = some_vector
    integer_scalar_var[:] = 5
    root_g.close()


def test_get_wall_adjacent_triangles():
    with OmegaHMesh(omega_h_mesh_to_convert) as mesh:
        wall_adjacent_triangles = mesh.get_wall_adjacent_triangles()
        assert len(wall_adjacent_triangles) > 0

        plasma_side = wall_adjacent_triangles[0::2]
        target_side = wall_adjacent_triangles[1::2]
        print("Wall adjacent triangles:")
        print("Plasma side:", plasma_side.tolist())
        print("Target side:", target_side.tolist())

    ref_plasma_side = np.array([0, 2, 3, 4, 10, 10, 11, 11], dtype=int)
    ref_target_side = np.array([12, 15, 17, 18, 21, 22, 24, 27], dtype=int)

    # sort plasma_side and target_side for comparison
    plasma_side = np.sort(plasma_side)
    target_side = np.sort(target_side)

    assert np.all(ref_plasma_side == plasma_side), (
        f"Plasma side cell id mismatch {ref_plasma_side=} != {plasma_side=}"
    )
    assert np.all(ref_target_side == target_side), (
        f"Target side cell id mismatch {ref_target_side=} != {target_side=}"
    )



import importlib.resources
import sys
from ctypes import CDLL

assert sys.platform == "linux"  # only works with linux

_filename = importlib.resources.files(__package__) / "lib/libomegah2csg.so"
_dll = CDLL(str(_filename))



import netCDF4
import numpy as np

from ..OmegaHMesh import OmegaHMesh

INT_UNUSED = 2000000000
DBL_UNUSED = 2.0e30
STR_UNUSED = "UNUSED                                                                                              "


# first one is on positive side
def sort_edge_to_face_map(edge_to_face_map, face2edge_map) -> np.ndarray:
    nedges = edge_to_face_map.shape[0]

    for i in range(0, nedges):
        first_face = edge_to_face_map[i, 0]
        for j in range(0, 3):
            edge = face2edge_map[first_face, 2 * j]
            if edge == i:
                first_face_sign = face2edge_map[first_face, 2 * j + 1]
                if first_face_sign == -1:  # then swap
                    edge_to_face_map[i, 0], edge_to_face_map[i, 1] = (
                        edge_to_face_map[i, 1],
                        edge_to_face_map[i, 0],
                    )
                    continue
                if first_face_sign == 1:
                    continue
                else:
                    raise RuntimeError(f"Sign cannon be {first_face_sign}")

    return edge_to_face_map


def write_nodes_to_file(node_coords, wallfile_name="wallfile.txt"):
    num_nodes = int(node_coords.shape[0] / 2)
    wallfile = open(wallfile_name, "w")
    wallfile.write("# Wallfile automatically generated by script\n")
    wallfile.write("1 \n")
    wallfile.write("#\n")
    wallfile.write("%d\n" % num_nodes)
    wallfile.write("#\n")
    for i in range(0, num_nodes):
        wallfile.write(
            "%26.20f   %26.20f \n" % (node_coords[2 * i], node_coords[2 * i + 1])
        )
    wallfile.close()


def write_dummy_bg_files_aux(nzone, nwall_input, stratum_start):
    nwallsegs = nwall_input
    nperline = 10

    f = open("plasmafile.txt", "w")
    f.write("%10s %10s %10s %10s %10s\n" % ("zone", "T(1)", "N(1)", "T(2)", "N(2)"))
    for i in range(0, nzone):
        f.write("%i %e %e %e %e\n" % (i + 1, 100.0, 1e19, 100.0, 1e19))
    f.close()

    f = open("sourcefile.txt", "w")
    f.write("#\nstratum\n#\n")
    for i in range(0, nwallsegs):
        f.write("%d " % (stratum_start + i))
        if ((i + 1) % nperline == 0) and i != nwallsegs - 1:
            f.write("\n")
    f.write("\n")
    f.write("#\nsegment\n#\n")
    for i in range(0, nwallsegs):
        f.write("%d " % (0))
        if ((i + 1) % nperline == 0) and i != nwallsegs - 1:
            f.write("\n")
    f.write("\n")
    f.write("#\nF\n#\n")
    for i in range(0, nwallsegs):
        f.write("%e " % (1e20))
        if ((i + 1) % nperline == 0) and i != nwallsegs - 1:
            f.write("\n")


def convert2degas2(
    mesh_filename, netcdf_filename="geometry.nc", create_aux_files=True, tol=1e-10
):
    """Convert Omega_h mesh to Degas2 NetCDF geometry file and auxiliary files.

    Parameters
    ----------
    mesh_filename : str
        Path to the Omega_h mesh file (.osh) to be converted. This mesh must have a boundary layer. See [workflow documentation](https://gist.github.com/Fuad-HH/ea6d4913adcfb4d857ef2a77aebaaa2c)
        for details on how to create such a mesh.
    netcdf_filename : str, optional
        Desired name for the output NetCDF file. Default is "geometry.nc".
    create_aux_files : bool, optional
        If True, creates auxiliary files (plasmafile.txt, sourcefile.txt, wallfile.txt) required for Degas2 simulations. Default is True.
    tol : float, optional
        Tolerance for numerical comparisons when determining edge orientations and other geometric properties. Default is 1e-10.

    """

    assert netcdf_filename.endswith(".nc"), (
        "Degas2 mesh name should end with .nc but given {}".format(netcdf_filename)
    )
    with OmegaHMesh(mesh_filename) as mesh:
        assert mesh.has_boundary_layer, (
            "Degas2 requires mesh to have a boundary layer. Use addBonudaryLayer tool from tomms."
        )
        [edge_coefficients, edge_types_, boundary_edge_ids, face2edge_map] = (
            mesh.get_all_geometry_info(tol=tol)
        )
        Ntri = face2edge_map.shape[0]  # ncells
        boundary_face_flag = mesh.get_boundary_face_flag()
        cell_bounding_boxes = mesh.get_cell_bounding_boxes()
        tri_volumes = mesh.get_cell_volumes()
        centroids = mesh.get_cell_centroids().reshape((Ntri, 2))
        edge_coordinates = mesh.get_edge_coordinates()
        edge_to_face_map = mesh.get_edge_to_face_map()
        # sorted like e0+, e0-, e1+, e1-, ...
        edge_to_face_map_sorted = sort_edge_to_face_map(edge_to_face_map, face2edge_map)
        wall_nodes_flag = mesh.get_integer_tag_array(0, "isOnWall")
        first_wall_adjacent_faces = mesh.get_wall_adjacent_triangles()
        first_wall_edges = mesh.get_wall_edge_ids()

        if create_aux_files:
            node_coordinates = mesh.get_node_coordinates()
            write_nodes_to_file(node_coordinates)

    num_boundary_face = boundary_face_flag.sum()
    Nplasma = Ntri - num_boundary_face
    Nedge = edge_coefficients.shape[0]
    Nwall = boundary_edge_ids.shape[0]
    # universal cells has Nwall boundaries and each triangle has 3 sides and 2 cut surfaces
    nboundaries = Nwall + 3 * Ntri
    nneighbors = Nedge * 2
    num_first_wall_points = np.sum(wall_nodes_flag)
    num_zones = Nplasma + num_first_wall_points
    nsectors = 2 * num_first_wall_points

    if create_aux_files:
        write_dummy_bg_files_aux(Nplasma, num_first_wall_points, Nplasma + 1)

    # FIXME NETCDF_CLASSIC is limited to 2GB
    root_g = netCDF4.Dataset(netcdf_filename, mode="w", format="NETCDF4_CLASSIC")

    # *********************************************************************************************** #
    # ---------------------------------- Dimensions ------------------------------------------------- #
    # *********************************************************************************************** #

    root_g.createDimension("vector", 3)
    root_g.createDimension("string", 300)
    root_g.createDimension("cell_info_ind", 4)
    root_g.createDimension("cell_ind", Ntri + 1)
    root_g.createDimension("surface_ind", Nedge)
    root_g.createDimension("boundary_ind", nboundaries)
    root_g.createDimension("neighbor_ind", nneighbors + 1)
    root_g.createDimension("neg_pos", 2)
    root_g.createDimension("surface_info_ind", 2)
    root_g.createDimension("surface_tx_ind", 2)
    root_g.createDimension("tx_ind_1", 3)
    root_g.createDimension("tx_ind_2", 4)
    root_g.createDimension("transform_ind", 1)
    root_g.createDimension("coeff_ind", 11)
    root_g.createDimension("zone_type_ind", 4)
    root_g.createDimension("zone_index_ind", 4)
    root_g.createDimension("zone_ind", num_zones + 1)
    root_g.createDimension("sector_ind", nsectors + 1)
    root_g.createDimension("sector_neg_pos_ind", 2)
    root_g.createDimension("sector_type_ind", 17)
    root_g.createDimension("vacuum_ind", 1)
    root_g.createDimension("plasma_ind", num_first_wall_points + 1)
    root_g.createDimension("target_ind", num_first_wall_points + 1)
    root_g.createDimension("wall_ind", 1)
    root_g.createDimension("exit_ind", 1)
    root_g.createDimension("sc_diag_name_string", 40)
    root_g.createDimension("diag_grp_ind", 4)
    root_g.createDimension("sc_diag_ind", 3 * num_first_wall_points)
    root_g.createDimension("de_symbol_string", 24)
    root_g.createDimension("de_name_string", 100)
    root_g.createDimension("de_grp_ind", 1)
    root_g.createDimension("de_zone_frags_ind", 100)
    root_g.createDimension("de_tot_view_ind", 1)
    root_g.createDimension("de_start_end_ind", 2)
    root_g.createDimension("de_view_ind", 1)
    print("Dimensions created. Now creating variables...")

    # *********************************************************************************************** #
    # ---------------------------------- Variables -------------------------------------------------- #
    # *********************************************************************************************** #

    ncells_var = root_g.createVariable("ncells", "i4")
    nsurfaces_var = root_g.createVariable("nsurfaces", "i4")
    nboundaries_var = root_g.createVariable("nboundaries", "i4")
    nneighbors_var = root_g.createVariable("nneighbors", "i4")
    ntransforms_var = root_g.createVariable("ntransforms", "i4")
    geometry_symmetry_var = root_g.createVariable("geometry_symmetry", "i4")
    universal_cell_min_var = root_g.createVariable(
        "universal_cell_min", "f8", ("vector",)
    )
    universal_cell_max_var = root_g.createVariable(
        "universal_cell_max", "f8", ("vector",)
    )
    universal_cell_vol_var = root_g.createVariable("universal_cell_vol", "f8")
    cells_var = root_g.createVariable(
        "cells",
        "i4",
        (
            "cell_ind",
            "cell_info_ind",
        ),
    )
    surfaces_var = root_g.createVariable(
        "surfaces",
        "i4",
        (
            "surface_ind",
            "surface_info_ind",
            "neg_pos",
        ),
    )
    surfaces_tx_ind_var = root_g.createVariable(
        "surfaces_tx_ind",
        "i4",
        (
            "surface_ind",
            "surface_tx_ind",
            "neg_pos",
        ),
    )
    surfaces_tx_mx_var = root_g.createVariable(
        "surfaces_tx_mx",
        "f8",
        (
            "transform_ind",
            "tx_ind_2",
            "tx_ind_1",
        ),
    )
    surface_sectors_var = root_g.createVariable(
        "surface_sectors",
        "i4",
        (
            "surface_ind",
            "surface_info_ind",
            "neg_pos",
        ),
    )
    boundaries_var = root_g.createVariable("boundaries", "i4", ("boundary_ind",))
    neighbors_var = root_g.createVariable("neighbors", "i4", ("neighbor_ind",))
    surface_coeffs_var = root_g.createVariable(
        "surface_coeffs",
        "f8",
        (
            "surface_ind",
            "coeff_ind",
        ),
    )
    surface_points_var = root_g.createVariable(
        "surface_points",
        "f8",
        (
            "surface_ind",
            "neg_pos",
            "vector",
        ),
    )
    zn_num_var = root_g.createVariable("zn_num", "i4")
    zone_type_num_var = root_g.createVariable("zone_type_num", "i4", ("zone_type_ind",))
    zone_type_var = root_g.createVariable("zone_type", "i4", ("zone_ind",))
    zone_index_var = root_g.createVariable(
        "zone_index",
        "i4",
        (
            "zone_ind",
            "zone_index_ind",
        ),
    )
    zone_index_min_var = root_g.createVariable(
        "zone_index_min", "i4", ("zone_index_ind",)
    )
    zone_index_max_var = root_g.createVariable(
        "zone_index_max", "i4", ("zone_index_ind",)
    )
    zone_pointer_var = root_g.createVariable("zone_pointer", "i4", ("zone_ind",))
    zone_volume_var = root_g.createVariable("zone_volume", "f8", ("zone_ind",))
    zone_center_var = root_g.createVariable(
        "zone_center",
        "f8",
        (
            "zone_ind",
            "vector",
        ),
    )
    zone_min_var = root_g.createVariable(
        "zone_min",
        "f8",
        (
            "zone_ind",
            "vector",
        ),
    )
    zone_max_var = root_g.createVariable(
        "zone_max",
        "f8",
        (
            "zone_ind",
            "vector",
        ),
    )
    nsectors_var = root_g.createVariable("nsectors", "i4")
    strata_var = root_g.createVariable("strata", "i4", ("sector_ind",))
    sector_strata_segment_var = root_g.createVariable(
        "sector_strata_segment", "i4", ("sector_ind",)
    )
    sectors_var = root_g.createVariable("sectors", "i4", ("sector_ind",))
    sector_zone_var = root_g.createVariable("sector_zone", "i4", ("sector_ind",))
    sector_surface_var = root_g.createVariable("sector_surface", "i4", ("sector_ind",))
    sector_points_var = root_g.createVariable(
        "sector_points",
        "f8",
        (
            "sector_ind",
            "sector_neg_pos_ind",
            "vector",
        ),
    )
    sector_type_pointer_var = root_g.createVariable(
        "sector_type_pointer",
        "i4",
        (
            "sector_ind",
            "sector_type_ind",
        ),
    )
    sc_vacuum_num_var = root_g.createVariable("sc_vacuum_num", "i4")
    vacuum_sector_var = root_g.createVariable("vacuum_sector", "i4", ("vacuum_ind",))
    sc_plasma_num_var = root_g.createVariable("sc_plasma_num", "i4")
    plasma_sector_var = root_g.createVariable("plasma_sector", "i4", ("plasma_ind",))
    sc_target_num_var = root_g.createVariable("sc_target_num", "i4")
    target_sector_var = root_g.createVariable("target_sector", "i4", ("target_ind",))
    target_material_var = root_g.createVariable(
        "target_material", "i4", ("target_ind",)
    )
    target_temperature_var = root_g.createVariable(
        "target_temperature", "f8", ("target_ind",)
    )
    target_recyc_coef_var = root_g.createVariable(
        "target_recyc_coef", "f8", ("target_ind",)
    )
    sc_wall_num_var = root_g.createVariable("sc_wall_num", "i4")
    wall_sector_var = root_g.createVariable("wall_sector", "i4", ("wall_ind",))
    wall_material_var = root_g.createVariable("wall_material", "i4", ("wall_ind",))
    wall_temperature_var = root_g.createVariable(
        "wall_temperature", "f8", ("wall_ind",)
    )
    wall_recyc_coef_var = root_g.createVariable("wall_recyc_coef", "f8", ("wall_ind",))
    sc_exit_num_var = root_g.createVariable("sc_exit_num", "i4")
    exit_sector_var = root_g.createVariable("exit_sector", "i4", ("exit_ind",))
    sc_diagnostic_grps_var = root_g.createVariable("sc_diagnostic_grps", "i4")
    sc_diag_max_bins_var = root_g.createVariable("sc_diag_max_bins", "i4")
    diagnostic_grp_name_var = root_g.createVariable(
        "diagnostic_grp_name",
        "c",
        (
            "diag_grp_ind",
            "sc_diag_name_string",
        ),
    )
    diagnostic_num_sectors_var = root_g.createVariable(
        "diagnostic_num_sectors", "i4", ("diag_grp_ind",)
    )
    diagnostic_var_var = root_g.createVariable(
        "diagnostic_var", "i4", ("diag_grp_ind",)
    )
    diagnostic_tab_index_var = root_g.createVariable(
        "diagnostic_tab_index", "i4", ("diag_grp_ind",)
    )
    diagnostic_min_var = root_g.createVariable(
        "diagnostic_min", "f8", ("diag_grp_ind",)
    )
    diagnostic_delta_var = root_g.createVariable(
        "diagnostic_delta", "f8", ("diag_grp_ind",)
    )
    diagnostic_spacing_var = root_g.createVariable(
        "diagnostic_spacing", "i4", ("diag_grp_ind",)
    )
    diagnostic_grp_base_var = root_g.createVariable(
        "diagnostic_grp_base", "i4", ("diag_grp_ind",)
    )
    sc_diag_size_var = root_g.createVariable("sc_diag_size", "i4")
    diagnostic_sector_tab_var = root_g.createVariable(
        "diagnostic_sector_tab", "i4", ("sc_diag_ind",)
    )
    de_grps_var = root_g.createVariable("de_grps", "i4")
    de_max_bins_var = root_g.createVariable("de_max_bins", "i4")
    de_zone_frags_dim_var = root_g.createVariable("de_zone_frags_dim", "i4")
    de_zone_frags_size_var = root_g.createVariable("de_zone_frags_size", "i4")
    detector_name_var = root_g.createVariable(
        "detector_name",
        "c",
        (
            "de_grp_ind",
            "de_name_string",
        ),
    )
    detector_num_views_var = root_g.createVariable(
        "detector_num_views", "i4", ("de_grp_ind",)
    )
    detector_var_var = root_g.createVariable("detector_var", "i4", ("de_grp_ind",))
    detector_tab_index_var = root_g.createVariable(
        "detector_tab_index", "i4", ("de_grp_ind",)
    )
    detector_min_var = root_g.createVariable("detector_min", "f8", ("de_grp_ind",))
    detector_delta_var = root_g.createVariable("detector_delta", "f8", ("de_grp_ind",))
    detector_spacing_var = root_g.createVariable(
        "detector_spacing", "i4", ("de_grp_ind",)
    )
    detector_total_views_var = root_g.createVariable("detector_total_views", "i4")
    de_view_points_var = root_g.createVariable(
        "de_view_points",
        "f8",
        (
            "de_tot_view_ind",
            "de_start_end_ind",
            "vector",
        ),
    )
    de_view_algorithm_var = root_g.createVariable(
        "de_view_algorithm", "i4", ("de_tot_view_ind",)
    )
    de_view_halfwidth_var = root_g.createVariable(
        "de_view_halfwidth", "f8", ("de_tot_view_ind",)
    )
    de_zone_frags_var = root_g.createVariable(
        "de_zone_frags", "f8", ("de_zone_frags_ind",)
    )
    de_zone_frags_start_var = root_g.createVariable(
        "de_zone_frags_start", "i4", ("de_tot_view_ind",)
    )
    de_zone_frags_num_var = root_g.createVariable(
        "de_zone_frags_num", "i4", ("de_tot_view_ind",)
    )
    de_zone_frags_zones_var = root_g.createVariable(
        "de_zone_frags_zones", "i4", ("de_zone_frags_ind",)
    )
    de_zone_frags_min_zn_var = root_g.createVariable(
        "de_zone_frags_min_zn", "i4", ("de_tot_view_ind",)
    )
    de_zone_frags_max_zn_var = root_g.createVariable(
        "de_zone_frags_max_zn", "i4", ("de_tot_view_ind",)
    )
    de_view_base_var = root_g.createVariable("de_view_base", "i4", ("de_grp_ind",))
    de_view_size_var = root_g.createVariable("de_view_size", "i4")
    de_view_tab_var = root_g.createVariable("de_view_tab", "i4", ("de_view_ind",))
    print("Variables created. Now filling the variables...")

    # *********************************************************************************************** #
    # ---------------------------------- Fill Up ---------------------------------------------------- #
    # *********************************************************************************************** #

    ncells_var[:] = Ntri
    nsurfaces_var[:] = Nedge
    nboundaries_var[:] = nboundaries
    nneighbors_var[:] = nneighbors
    ntransforms_var[:] = 0  # hardcoded to zero
    geometry_symmetry_var[:] = 2  # hardcoded to two

    assert cell_bounding_boxes.size == Ntri * 4
    universal_cell_min = [
        np.min(cell_bounding_boxes[0::4]),
        0.0,
        np.min(cell_bounding_boxes[1::4]),
    ]
    universal_cell_max = [
        np.max(cell_bounding_boxes[2::4]),
        6.28318530717959,
        np.max(cell_bounding_boxes[3::4]),
    ]
    universal_cell_min_var[:] = universal_cell_min
    universal_cell_max_var[:] = universal_cell_max

    total_volume = tri_volumes.sum()
    universal_cell_vol_var[:] = total_volume

    # ------------------------- Cells -------------------------------------- #
    # in note ncells = Ntri = Nplasma + 2*Nwall; ncells = ncells = Ntri+2*Nwall
    cells = -1 + np.zeros([Ntri + 1, 4], dtype=int)  # ask
    cells[0, 0:4] = [1, Nwall, Nwall, 0]

    cells[1 : Ntri + 1, 0] = 1 + Nwall + 3 * np.array(range(0, Ntri), dtype=int)
    cells[1 : Ntri + 1, 1] = 3
    cells[1 : Ntri + 1, 2] = 3

    cells[1 : Nplasma + 1, 3] = np.array(range(1, Nplasma + 1), dtype=int)
    # now comes the wall zones
    for i, face in enumerate(first_wall_adjacent_faces[1::2]):
        cells[face + 1, 3] = i + Nplasma + 1

    assert np.max(cells[:, 3]) == num_zones, (
        f"Error: Upto this point, {np.max(cells[:, 3])=} != {num_zones=}"
    )

    # all remaining cells has zone_id = num_zones
    # the remaining values should be -1
    cells[cells[:, 3] == -1, 3] = num_zones + 1

    cells_var[:] = cells

    # ------------------------- Boundaries ------------------------------------ #
    boundaries = np.zeros(nboundaries, dtype=int)
    for i in range(0, Ntri):
        bdy_start = Nwall + i * 3

        boundaries[bdy_start : bdy_start + 3] = [
            (face2edge_map[i, 0] + 1) * face2edge_map[i, 1],
            (face2edge_map[i, 2] + 1) * face2edge_map[i, 3],
            (face2edge_map[i, 4] + 1) * face2edge_map[i, 5],
        ]

    # fill edges of the universal cell
    boundaries[0:Nwall] = np.sign(boundary_edge_ids) * (np.abs(boundary_edge_ids) + 1)

    boundaries_var[:] = boundaries

    # ------------------------ Surfaces ------------------------------------ #
    surfaces = np.zeros([Nedge, 2, 2], dtype=int)
    # pointers uses 0 based indexing
    surfaces[:Nedge, 0, 0] = 1 + 2 * np.arange(Nedge)
    surfaces[:Nedge, 0, 1] = 0 + 2 * np.arange(Nedge)
    # number of faces
    for i in range(0, Nedge):
        pos_face = edge_to_face_map_sorted[i, 0]
        neg_face = edge_to_face_map_sorted[i, 1]
        if neg_face == -1 and pos_face != -1:
            surfaces[i, 1, 0] = 0
            surfaces[i, 1, 1] = 1
        elif pos_face == -1 and neg_face != -1:
            surfaces[i, 1, 1] = 0
            surfaces[i, 1, 0] = 1
        elif neg_face != -1 and pos_face != -1:
            surfaces[i, 1, :] = 1
        else:
            raise RuntimeError(
                f"pos_face={pos_face}, neg_face={neg_face} is not correct!"
            )

    surfaces_var[:] = surfaces

    # ----------------------- Neighbors (Edge Adjacency Info) ----------------- #
    neighbors = np.zeros(nneighbors + 1, dtype=int)
    edge_to_face_map_sorted[~(edge_to_face_map_sorted == -1)] += (
        1  # use one based indexing
    )
    neighbors[0 : Nedge * 2] = edge_to_face_map_sorted.reshape(Nedge * 2)
    neighbors_var[:] = neighbors

    # --------------------- Surface Coefficients ------------------------------ #
    # TODO Create a better documentation of the coeffs
    surface_coeffs = np.zeros([Nedge, 11])
    # -b ^ 2 for a cone, -R ^ 2 for a cylinder, -Z0 for a plane
    surface_coeffs[:Nedge, 0] = edge_coefficients[:, 3]
    # b = 0 for a cylinder, 1/2 for a plane, and intercept for a cone
    surface_coeffs[:Nedge, 3] = edge_coefficients[:, 2]
    # m=1 for cylinder, slope for a cone, and 0 for a plane
    surface_coeffs[:Nedge, 4] = edge_coefficients[:, 0]
    surface_coeffs[:Nedge, 5] = edge_coefficients[:, 0]
    # in notebook np.where(np.logical_or(planecond,cylcond),np.zeros(Nsurfs),-np.ones(Nsurfs))
    surface_coeffs[:Nedge, 6] = edge_coefficients[:, 1]
    # cone up or down direction flag
    surface_coeffs[:Nedge, 10] = edge_coefficients[:, 4]

    surface_coeffs_var[:] = surface_coeffs

    # --------------------- Zones ------------------------------------------ #
    zone_center = np.zeros([num_zones + 1, 3])
    zone_center[0:Nplasma, 0] = centroids[0:Nplasma, 0]
    zone_center[0:Nplasma, 2] = centroids[0:Nplasma, 1]
    # they do specify it with the first cells centroid we are doing it with the last
    # for this zone
    assert cells[-1, 3] == num_zones + 1, (
        f"The next 2 lines won't work if this assertion fails. {cells[-1,3]=}, {num_zones=}"
    )
    zone_center[-1, 0] = centroids[-1, 0]
    zone_center[-1, 2] = centroids[-1, 1]
    # wall adjacent zones
    outside_wall_faces = first_wall_adjacent_faces[1::2]
    zone_center[Nplasma:-1, 0] = centroids[outside_wall_faces, 0]
    zone_center[Nplasma:-1, 2] = centroids[outside_wall_faces, 1]

    zone_center_var[:] = zone_center

    zone_min = np.zeros((num_zones + 1, 3))
    zone_min[0:Nplasma, 0] = cell_bounding_boxes[0 : Nplasma * 4 : 4]
    zone_min[0:Nplasma, 2] = cell_bounding_boxes[1 : Nplasma * 4 : 4]

    zone_min[Nplasma:-1, 0] = cell_bounding_boxes[outside_wall_faces * 4]
    zone_min[Nplasma:-1, 2] = cell_bounding_boxes[outside_wall_faces * 4 + 1]

    zone_min[-1, :] = [universal_cell_min[0], 0.0, universal_cell_min[2]]
    zone_min_var[:] = zone_min

    zone_max = np.zeros((num_zones + 1, 3))
    zone_max[0:Nplasma, 0] = cell_bounding_boxes[2 : Nplasma * 4 : 4]
    zone_max[0:Nplasma, 2] = cell_bounding_boxes[3 : Nplasma * 4 : 4]

    zone_max[Nplasma:-1, 0] = cell_bounding_boxes[outside_wall_faces * 4 + 2]
    zone_max[Nplasma:-1, 2] = cell_bounding_boxes[outside_wall_faces * 4 + 3]

    zone_max[-1, :] = [universal_cell_max[0], 0.0, universal_cell_max[2]]
    zone_max_var[:] = zone_max

    zone_volume = np.empty(num_zones + 1)
    zone_volume[0:Nplasma] = tri_volumes[0:Nplasma]

    zone_volume[Nplasma:-1] = tri_volumes[outside_wall_faces]

    # do not move this line here in this block
    zone_volume[-1] = total_volume - np.sum(zone_volume)
    zone_volume_var[:] = zone_volume

    zone_type = 2 * np.ones(num_zones + 1, dtype=int)
    zone_type[Nplasma:] = 3
    zone_type_var[:] = zone_type

    zone_pointer = np.array(range(1, num_zones + 2))
    zone_pointer[Nplasma:] = np.array(range(1, num_zones - Nplasma + 2))
    zone_pointer_var[:] = zone_pointer

    # ----------------------- Surface Points ------------------------------ #
    surface_points = np.zeros([Nedge, 2, 3])
    surface_points[0:Nedge, :, 0] = edge_coordinates[0 : Nedge * 4 : 2].reshape(
        (Nedge, 2)
    )
    surface_points[0:Nedge, :, 2] = edge_coordinates[1 : Nedge * 4 : 2].reshape(
        (Nedge, 2)
    )
    surface_points_var[:] = surface_points

    # -------------------------- Sector ---------------------------------------- #

    sc_plasma_num = num_first_wall_points
    sc_target_num = num_first_wall_points
    plasma_sector = np.zeros(sc_plasma_num + 1, dtype=int)
    target_sector = np.zeros(sc_target_num + 1, dtype=int)
    plasma_sector[0] = INT_UNUSED
    target_sector[0] = INT_UNUSED
    plasma_sector[1:] = 1.0 + 2 * np.arange(sc_plasma_num)
    target_sector[1:] = 2.0 + 2 * np.arange(sc_target_num)
    plasma_sector_var[:] = plasma_sector
    target_sector_var[:] = target_sector

    sectors = np.empty([nsectors + 1], dtype=int)
    sectors[0] = INT_UNUSED
    sector_surface = np.zeros(nsectors + 1, dtype=int)
    sector_surface[0] = INT_UNUSED
    for i in range(0, first_wall_edges.shape[0]):
        wall_edge = first_wall_edges[i]
        plasma_cell = first_wall_adjacent_faces[2 * i + 0]
        wall_cell = first_wall_adjacent_faces[2 * i + 1]
        positive_cell = edge_to_face_map_sorted[wall_edge, 0] - 1
        if plasma_cell == positive_cell:
            sector_surface[2 * i + 1] = wall_edge + 1
            sector_surface[2 * i + 2] = -wall_edge - 1
            sectors[2 * i + 1] = i * 2 + 2
            sectors[2 * i + 2] = i * 2 + 1
        elif wall_cell == positive_cell:
            sector_surface[2 * i + 1] = -wall_edge - 1
            sector_surface[2 * i + 2] = wall_edge + 1
            sectors[2 * i + 1] = i * 2 + 1
            sectors[2 * i + 2] = i * 2 + 2
        else:
            raise RuntimeError(
                f"Incorrect cells {wall_edge=}, {plasma_cell=}, {wall_cell=}, {positive_cell=} "
            )

    sector_surface_var[:] = sector_surface
    sectors_var[:] = sectors

    surface_sectors = np.zeros([Nedge, 2, 2], dtype=int)

    # the real edges (not cut surfaces) will have sectors associated
    current_sector_pointer = 1
    for i in range(0, Nedge):
        # fixme replace this isin for performance
        if not np.isin(i, first_wall_edges):
            surface_sectors[i, 1, :] = 0
            surface_sectors[i, 0, :] = current_sector_pointer
        else:
            surface_sectors[i, 1, :] = 1
            surface_sectors[i, 0, 0] = current_sector_pointer
            surface_sectors[i, 0, 1] = current_sector_pointer + 1
            current_sector_pointer += 2

    surface_sectors_var[:] = surface_sectors

    sector_points = np.zeros([nsectors + 1, 2, 3])
    sector_points[0, :, :] = DBL_UNUSED
    for i in range(0, num_first_wall_points):
        wall_edge = first_wall_edges[i]
        sector_points[2 * i + 1, 0, :] = [
            edge_coordinates[wall_edge * 4 + 0],
            0.0,
            edge_coordinates[wall_edge * 4 + 1],
        ]
        sector_points[2 * i + 1, 1, :] = [
            edge_coordinates[wall_edge * 4 + 2],
            0.0,
            edge_coordinates[wall_edge * 4 + 3],
        ]
        sector_points[2 * i + 2, :, :] = sector_points[2 * i + 1, :, :]

    sector_points_var[:] = sector_points

    sector_zone = np.zeros(nsectors + 1, dtype=int)
    sector_zone[0] = INT_UNUSED
    sector_zone[1:] = cells[first_wall_adjacent_faces + 1, 3]
    sector_zone_var[:] = sector_zone

    sector_type_pointer = INT_UNUSED * np.ones([nsectors + 1, 17], dtype=int)
    sc_diag_size = 3 * num_first_wall_points
    diagnostic_sector_tab = np.zeros(sc_diag_size, dtype=int)

    sector_type_pointer[0, :] = INT_UNUSED / 2
    # fixme after surface, face, etc.
    for i in range(0, num_first_wall_points):
        sector_type_pointer[2 * i + 1, 1] = i + 1
        sector_type_pointer[2 * i + 2, 2] = i + 1
        sector_type_pointer[2 * i + 2, 5:8] = i + 1

        diagnostic_sector_tab[i] = 2 * i + 2
        diagnostic_sector_tab[num_first_wall_points + i] = 2 * i + 2
        diagnostic_sector_tab[2 * num_first_wall_points + i] = 2 * i + 2

    sector_type_pointer_var[:] = sector_type_pointer
    diagnostic_sector_tab_var[:] = diagnostic_sector_tab

    # collection of signed edges, each cells are bounded by its strata
    # sector is surface and a direction, something physically happennign
    # for each sector, what strata is associated with
    # strata index and which edge is it

    sector_strata_segment = np.zeros(nsectors + 1, dtype=int)
    sector_strata_segment[0] = INT_UNUSED
    # can be wrong since difference in boundary layer
    sector_strata_segment_var[:] = sector_strata_segment

    # ----------------------- Unmodified Variables ----------------------------- #
    de_grps = 0
    de_view_size = 1
    de_grps = 0
    de_max_bins = 0
    de_zone_frags_size = 0
    de_zone_frags_ind = 100
    detector_total_views = 0
    sc_vacuum_num = 0
    sc_plasma_num = num_first_wall_points
    sc_target_num = num_first_wall_points
    sc_wall_num = 0
    sc_exit_num = 0
    sc_diagnostic_grps = 3
    sc_diag_max_bins = 4

    zn_num_var[:] = num_zones + 1
    vacuum_sector = np.zeros(sc_vacuum_num + 1, dtype=int)
    vacuum_sector[0] = INT_UNUSED
    de_view_tab = INT_UNUSED * np.ones(de_view_size, dtype=int)
    de_view_tab_var[:] = de_view_tab
    de_view_size_var[:] = de_view_size
    de_view_base = INT_UNUSED * np.ones(de_grps + 1, dtype=int)
    de_view_base_var[:] = de_view_base
    sc_plasma_num_var[:] = sc_plasma_num
    sc_target_num_var[:] = sc_target_num
    sc_wall_num_var[:] = sc_wall_num
    sc_exit_num_var[:] = sc_exit_num
    sc_diagnostic_grps_var[:] = sc_diagnostic_grps
    sc_diag_max_bins_var[:] = sc_diag_max_bins
    sc_vacuum_num_var[:] = sc_vacuum_num

    # target_temperature = np.zeros(sc_target_num+1) # ask
    Twall = 300 * 1.380649e-23
    target_temperature = Twall * np.ones(sc_target_num + 1)
    target_temperature[0] = DBL_UNUSED
    target_temperature_var[:] = target_temperature

    wall_recyc_coef = np.zeros(sc_wall_num + 1)
    wall_recyc_coef[0] = DBL_UNUSED
    wall_recyc_coef_var[:] = wall_recyc_coef

    exit_sector = np.zeros(sc_exit_num + 1, dtype=int)
    exit_sector[0] = INT_UNUSED
    exit_sector_var[:] = exit_sector

    vacuum_sector = np.zeros(sc_vacuum_num + 1, dtype=int)
    wall_sector = np.zeros(sc_wall_num + 1, dtype=int)
    wall_material = np.zeros(sc_wall_num + 1, dtype=int)
    wall_temperature = np.zeros(sc_wall_num + 1)
    wall_recyc_coef = np.zeros(sc_wall_num + 1)
    vacuum_sector[0] = INT_UNUSED
    wall_sector[0] = INT_UNUSED
    wall_material[0] = INT_UNUSED
    wall_temperature[0] = DBL_UNUSED
    wall_recyc_coef[0] = DBL_UNUSED
    vacuum_sector_var[:] = vacuum_sector
    wall_sector_var[:] = wall_sector
    wall_material_var[:] = wall_material
    wall_temperature_var[:] = wall_temperature

    recyc_coef = 0.99
    target_recyc_coef = recyc_coef * np.ones(sc_target_num + 1)
    target_recyc_coef[0] = DBL_UNUSED
    target_recyc_coef_var[:] = target_recyc_coef

    target_material = 6 * np.ones(sc_target_num + 1, dtype=int)
    target_material[0] = INT_UNUSED
    target_material_var[:] = target_material

    zone_type_num = np.zeros(4, dtype=int)
    zone_type_num[1] = Nplasma
    zone_type_num[2] = num_zones - Nplasma + 1
    zone_type_num_var[:] = zone_type_num

    surfaces_tx_ind = np.zeros([Nedge, 2, 2], dtype=int)
    surfaces_tx_mx = DBL_UNUSED * np.ones([1, 4, 3])
    surfaces_tx_ind_var[:] = surfaces_tx_ind
    surfaces_tx_mx_var[:] = surfaces_tx_mx

    zone_index = np.zeros([num_zones + 1, 4], dtype=int)
    zone_index[:, 3] = np.array(range(1, num_zones + 2))
    zone_index_var[:] = zone_index

    zone_index_min = np.zeros(4, dtype=int)
    zone_index_max = np.zeros(4, dtype=int)
    zone_index_min_var[:] = zone_index_min
    zone_index_max_var[:] = zone_index_max

    nsectors_var[:] = nsectors

    strata = np.empty(nsectors + 1)
    strata[0] = INT_UNUSED
    # can be wrong since I didn't find exact definition.
    strata[1:] = np.repeat(
        np.arange(Nplasma + 1, Nplasma + 1 + num_first_wall_points), 2
    )
    strata_var[:] = strata

    detector_num_views = INT_UNUSED * np.ones(shape=de_grps + 1, dtype=int)
    detector_var = INT_UNUSED * np.ones(de_grps + 1, dtype=int)
    detector_tab_index = INT_UNUSED * np.ones(de_grps + 1, dtype=int)
    detector_min = DBL_UNUSED * np.ones(de_grps + 1)
    detector_delta = DBL_UNUSED * np.ones(de_grps + 1)
    detector_spacing = INT_UNUSED * np.ones(de_grps + 1, dtype=int)
    de_view_points = DBL_UNUSED * np.ones([detector_total_views + 1, 2, 3])
    de_view_algorithm = INT_UNUSED * np.ones(detector_total_views + 1, dtype=int)
    de_view_halfwidth = DBL_UNUSED * np.ones(detector_total_views + 1)
    de_zone_frags = np.zeros(de_zone_frags_ind)
    de_zone_frags_start = np.zeros(detector_total_views + 1, dtype=int)
    de_zone_frags_num = np.zeros(detector_total_views + 1, dtype=int)
    de_zone_frags_zones = np.zeros(de_zone_frags_ind)
    de_zone_frags_zones[:] = 4
    de_zone_frags_zones[0] = INT_UNUSED
    de_zone_frags_min_zn = np.zeros(detector_total_views + 1, dtype=int)
    de_zone_frags_max_zn = np.zeros(detector_total_views + 1, dtype=int)
    detector_num_views_var[:] = detector_num_views
    detector_var_var[:] = detector_var
    detector_tab_index_var[:] = detector_tab_index
    detector_min_var[:] = detector_min
    detector_delta_var[:] = detector_delta
    detector_spacing_var[:] = detector_spacing
    detector_total_views_var[:] = detector_total_views
    de_view_points_var[:] = de_view_points
    de_view_algorithm_var[:] = de_view_algorithm
    de_view_halfwidth_var[:] = de_view_halfwidth
    de_zone_frags_var[:] = de_zone_frags
    de_zone_frags_start_var[:] = de_zone_frags_start
    de_zone_frags_num_var[:] = de_zone_frags_num
    de_zone_frags_zones_var[:] = de_zone_frags_zones
    de_zone_frags_min_zn_var[:] = de_zone_frags_min_zn
    de_zone_frags_max_zn_var[:] = de_zone_frags_max_zn
    diagnostic_num_sectors = np.array([INT_UNUSED, Nwall, Nwall, Nwall])
    diagnostic_var = np.array([INT_UNUSED, 0, 1, 2])
    diagnostic_tab_index = np.array([INT_UNUSED, 0, 4, 4])
    diagnostic_spacing = np.array([INT_UNUSED, 0, 2, 1])
    diagnostic_grp_base = np.array([INT_UNUSED, 0, Nwall, 2 * Nwall])
    diagnostic_min = np.array([DBL_UNUSED, 0, -45.5803383244769, 0.174532925199433])
    diagnostic_delta = np.array([DBL_UNUSED, 0, 2.30258509299405, 0.349065850398866])
    diagnostic_num_sectors_var[:] = diagnostic_num_sectors
    diagnostic_var_var[:] = diagnostic_var
    diagnostic_tab_index_var[:] = diagnostic_tab_index
    diagnostic_min_var[:] = diagnostic_min
    diagnostic_delta_var[:] = diagnostic_delta
    diagnostic_spacing_var[:] = diagnostic_spacing
    diagnostic_grp_base_var[:] = diagnostic_grp_base
    sc_diag_size_var[:] = sc_diag_size
    de_grps_var[:] = de_grps
    de_max_bins_var[:] = de_max_bins
    de_zone_frags_dim = 100
    de_zone_frags_dim_var[:] = de_zone_frags_dim
    de_zone_frags_size_var[:] = de_zone_frags_size

    diagnostic_grp_name_var[0, :] = "UNUSED                                  "
    diagnostic_grp_name_var[1, :] = "Wall and Target Counts                  "
    diagnostic_grp_name_var[2, :] = "Wall and Target Energy Spectrum         "
    diagnostic_grp_name_var[3, :] = "Wall and Target Angle Spectrum          "
    detector_name_var[:] = (
        "UNUSED                                                                                              "
    )

    root_g.sync()
    root_g.close()
